#!/bin/bash

# Copyright (c) 2022 xFusion Digital Technologies Co., Ltd. All rights reserved.
# Description: parse klp patch information
# Author: wuqiong
# Create: 2022-01-22

declare -A KLP_MOD_FUNCS_DICT
declare -A KLP_OFFSET_STRING_DICT

declare -a KLP_FUNCS
declare -a KLP_FUNC_MOD_OFFSETS

function parse_klp_strings() {
    local klp_module=$1
    local random="${RANDOM}"
    if [ -z "${random}" ]; then
        random=`date +%s`
    fi

    local tmp_dir="/tmp/parse_klp.${random}"
    mkdir -m 700 -p ${tmp_dir}

    local kpatch_strings_file=`basename ${klp_module}`
    kpatch_strings_file="${tmp_dir}/${kpatch_strings_file}.strings"

    local kpatch_strings=`readelf -p .kpatch.strings ${klp_module}|grep '\[.*\] .*'|awk '{print $NF}'`

    KLP_OFFSET_STRING_DICT=()
    local offset=0
    for string in ${kpatch_strings}
    do
        KLP_OFFSET_STRING_DICT[${offset}]=${string}
        ((offset=${offset}+${#string}+1))
    done

    echo "--------------------------------------------------------------"
    echo "parse_klp_strings:"
    echo "--------------------------------------------------------------"
    for offset in $(echo ${!KLP_OFFSET_STRING_DICT[*]})
    do
        echo "${offset} : ${KLP_OFFSET_STRING_DICT[${offset}]}"
    done
    echo "--------------------------------------------------------------"

    rm -Rf ${tmp_dir}
}

function parse_klp_funcs_and_mod_offsets() {
    local klp_module=$1
    local entry=`readelf -r "${klp_module}" |grep '.rela.kpatch.funcs'|awk '{print $(NF-1)}'`
    ((entry=$entry+1))
    local funcs=`readelf -r "${klp_module}" |grep -A $entry '.rela.kpatch.funcs' | grep -v '.kpatch.strings' | awk '{print $(NF-2)}' | tail -n +3`
    local offsets=`readelf -r "${klp_module}" |grep -A $entry '.rela.kpatch.funcs' | grep '.kpatch.strings' | awk '{print $NF}'`

    KLP_FUNCS=()
    for func in ${funcs}
    do
        KLP_FUNCS=(${KLP_FUNCS[@]} ${func})
    done

    KLP_FUNC_MOD_OFFSETS=()
    local index=1
    for offset in ${offsets}
    do
        if [ `expr ${index} % 2` = 0 ]; then
            KLP_FUNC_MOD_OFFSETS[${index}-1]=0x${offset}
        fi
        ((index=$index+1))
    done

    echo "--------------------------------------------------------------"
    echo "parse_klp_funcs_and_mod_offsets:"
    echo "--------------------------------------------------------------"
    echo "funcs: ${KLP_FUNCS[@]}"
    echo "mods: ${KLP_FUNC_MOD_OFFSETS[@]}"
    echo "--------------------------------------------------------------"
}

function gen_klp_mod_func_dict() {
    local index=0
    KLP_MOD_FUNCS_DICT=()
    for offset in ${KLP_FUNC_MOD_OFFSETS[@]}
    do
        ((num=${offset}))
        local mod=${KLP_OFFSET_STRING_DICT[${num}]}
        local func=${KLP_FUNCS[$index]}
        echo "${mod}:${func}"

	declare -a funcs
        funcs=${KLP_MOD_FUNCS_DICT[${mod}]}
        if [ -z "${funcs}" ]; then
            funcs=()
        fi
        funcs=(${funcs[@]} ${func})

        KLP_MOD_FUNCS_DICT[${mod}]=${funcs[@]}

        ((index=$index+1))
    done

    echo "--------------------------------------------------------------"
    echo "gen_klp_mod_func_dict:"
    echo "--------------------------------------------------------------"
    for mod in $(echo ${!KLP_MOD_FUNCS_DICT[*]})
    do
        echo "${mod} : ${KLP_MOD_FUNCS_DICT[${mod}]}"
    done
    echo "--------------------------------------------------------------"
}

function klp_parse() {
    klp_module=$1
    if [ ! -e ${klp_module} ]; then
        echo "Cannot find \"${klp_module}\""
        return 2
    fi

    parse_klp_strings ${klp_module}
    parse_klp_funcs_and_mod_offsets ${klp_module}
    gen_klp_mod_func_dict

    return 0
}
